/**
 * @file  DFFileDevice.cpp
 * @brief Implementierung der Klasse CDFFileDevice.
 */
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "DFFileDevice.h"
#include <time.h>

CDFFileDevice::CDFFileDevice()
{
	// Initialisieren
	Reset();
	// Fehlerobjekt zuruecksetzen
	ResetError();
}

CDFFileDevice::~CDFFileDevice()
{
}

void CDFFileDevice::Reset()
{
	// Basis Reset
	CDFFileIntelHex::Reset();

	// Dateiname der geladenen Datei halten
	memset(m_fileName, 0, sizeof(m_fileName));

	// Member initialisieren
	m_numSerial = 0;

	// Header loeschen
	memset(m_headBuf, 0, sizeof(m_headBuf));
	m_header = (SDFFileDeviceHeader*)m_headBuf;

	// Cryptpuffer loeschen
	m_isCrypt = 0;
	memset(m_crypt, 0xFF, sizeof(m_crypt));

	// Firmwaredaten loeschen
	m_fileRawFirmware.Reset();
	// Anzeigedaten loeschen
	m_fileDisplay.Reset();
	// Geraetetexte loeschen
	m_fileDeviceText.Reset();
}

SDFFileDeviceHeader CDFFileDevice::GetHeader(char *fileName)
{
	// Lesen der Hexdatei
	FILE *fp;
	char data[0x400];

	unsigned char headerBuf[IV_DFFILEDEVICE_SIZE_HEADER];
	memset(&headerBuf, 0, sizeof(headerBuf));
	SDFFileDeviceHeader *header = (SDFFileDeviceHeader*)headerBuf;

	if ((fp = fopen(fileName, "r")) == (FILE*)0) {
		// Oeffnen der Datei gescheitert
		m_errorState = ES_FILE_OPEN;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_FILE_OPEN_TEXT);
		return *header;
	}

	unsigned long numLine = 0;
	unsigned char count;
	unsigned short i, address;

	while(!feof(fp) && (m_errorState == ES_NO)) {
	
		if (fgets(data, 0x400, fp) != 0) {
			
			switch (GetType(data)) {
			case INTELHEX_TYPE_DAT: // Data record
			case INTELHEX_TYPE_EOF: // End-of-file record
			case INTELHEX_TYPE_ESA: // Extended Segment Address Record
			case INTELHEX_TYPE_SSA: // Start Segment Address Record
			case INTELHEX_TYPE_ELA: // Extended Linear Address Record
			case INTELHEX_TYPE_SLA: // Start Linear Address Record
				break;

			case INTELHEX_TYPE_DEVICE_HEADER:

				address = GetAddress(data);
				count   = GetCount(data);
				// spaetere Groessenanderung vorbehalten!
				for (i=address; (i < IV_DFFILEDEVICE_SIZE_HEADER) && ((i - address) < count); i++) {

					headerBuf[i] = GetData8(data, i - address);
				}
				break;

			case INTELHEX_TYPE_DEVICE_EEPROM:
			case INTELHEX_TYPE_DEVICE_FLASH:
				break;
			}

			numLine++;
		}
	}
	fclose(fp);	

	// Abgeaenderter Header ab 03.01.33 ist groesser 4 Byte
	if (header->sizeData) { // Vorher war nur die Version vorhanden!

		// CRC16 von Header pruefen!
		// Der crcHead ist immer der letzte 16bit Wert in der Struktur und
		// geht ueber den gesamten Dateiheader, sich selbst ausgeschlossen.
		if (*(unsigned short*)&headerBuf[header->sizeData - sizeof(unsigned short)] != GetCRC16(headerBuf, header->sizeData - sizeof(unsigned short))) {

			// Fehler zuweisen
			m_errorState = ES_HEAD_CRC;
			sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_HEAD_CRC_TEXT);

			memset(&headerBuf, 0, sizeof(headerBuf));
			return *header;
		}

		// Aeltere Struktur auf aktuell verwendete Version anpassen.
		if (header->sizeData < sizeof(SDFFileDeviceHeader)) {

			//############################################
			// SoftwareInfo ggf. auf die aktuelle Version anpassen.
			header->softwareInfo = CheckSoftwareInfo(header->softwareInfo);

			// CRC nach hinten verschieben, jedoch die Struktur-Version (sizeData) beibehalten.
			*(unsigned short*)&headerBuf[header->sizeData - sizeof(unsigned short)] = 0;
			header->crcHead = GetCRC16(headerBuf, header->sizeData - sizeof(unsigned short));

		} else if (header->sizeData > sizeof(SDFFileDeviceHeader)) {

			// Fehler zuweisen
			m_errorState = ES_VERSION; // Die Dateiversion kann nicht verarbeitet werden, da ungewisse Daten vorliegen.
			sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_VERSION_TEXT);

			memset(&headerBuf, 0, sizeof(headerBuf));
			return *header;
		}

	} else {

		// Fehler zuweisen
		m_errorState = ES_FUNCTION_UNAVAILABLE;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_FUNCTION_UNAVAILABLE_TEXT);

		memset(&headerBuf, 0, sizeof(headerBuf));
		return *header;
	}

	return *header;
}

int CDFFileDevice::CheckVersionFile(SDFFileDeviceHeader &header)
{
	// Version auswerten (Majornumber muss uebereinstimmen) Erst ab: 03.01.33
	if (((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_PZE   & 0xFFFF0000)) &&
		((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_ZK    & 0xFFFF0000)) &&
		((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_BDE   & 0xFFFF0000)) &&
		((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_MOBIL & 0xFFFF0000)) &&
		((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_MDE   & 0xFFFF0000)) &&
		((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_AE        & 0xFFFF0000)) &&
		((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_TIMEBOY   & 0xFFFF0000)) &&
		((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_MULTIPLE   & 0xFFFF0000)) ) {

		// Fehler zuweisen
		m_errorState = ES_DEVICE_UNKNOWN;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_DEVICE_UNKNOWN_TEXT);
		return FALSE;
	}

	return TRUE;
}

unsigned long CDFFileDevice::GetVersionFile(char *fileName)
{
	SDFFileDeviceHeader header = GetHeader(fileName);

	if (!header.sizeData || !CheckVersionFile(header)) {

		return FALSE;
	}

	return header.versionFile;
}

unsigned long CDFFileDevice::GetVersionFirmware(char *fileName, char *versionFirmware)
{
	SDFFileDeviceHeader header = GetHeader(fileName);

	if (!header.sizeData || !CheckVersionFile(header)) {
	
		return FALSE;
	}

	return sprintf(versionFirmware, "%s", header.versionFirmware);
}

int CDFFileDevice::Load(char *fileName) 
{
	// Umgebungsreset
	Reset();
	// Fehlerobjekt zuruecksetzen
	ResetError();

	// Header lesen
	*m_header = GetHeader(fileName);

	if (!m_header->sizeData || !CheckVersionFile(*m_header)) {

		// Header loeschen
		memset(m_headBuf, 0, sizeof(m_headBuf));
		m_header = (SDFFileDeviceHeader*)m_headBuf;
		return FALSE;
	}
	
	// Dateiname halten
	strcpy(m_fileName, fileName);

	// Firmwaredaten laden
	if (!m_fileRawFirmware.Load(fileName)) {

		return FALSE;
	}

	// CRC32 von Program-Flash pruefen
	if (m_header->crcProgram != GetCRC32(GetController()->GetMemProgram(0), GetController()->GetProgramLastAddress() - GetController()->GetProgramFirstAddress() + 1)) {

		// Fehler zuweisen
		m_errorState = ES_PROGRAM_CRC;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_PROGRAM_CRC_TEXT);
		return FALSE;
	}

	// Displaydaten laden
	if (!m_fileDisplay.Load(fileName)) {

		// Keine Displaydaten vorhanden?
		if (ES_NO_DATA != m_fileDisplay.m_errorState) {

			m_errorState = ES_LOAD_DISPLAY_DATA;
			sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_LOAD_DISPLAY_DATA_TEXT);
			return FALSE;
		}
	}
	// Geraetetexte laden
	if (!m_fileDeviceText.Load(fileName)) {

		// Keine Textdaten vorhanden?
		if (ES_NO_DATA != m_fileDeviceText.GetLastErrorState()) {

			m_errorState = ES_LOAD_DEVICE_TEXT;
			sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_LOAD_DEVICE_TEXT_TEXT);
			return FALSE;
		}
	}

	return TRUE;
}

int CDFFileDevice::Save(char *fileName, SDFFileDeviceHeader &header) 
{
	// Fehlerobjekt zuruecksetzen
	ResetError();

	//############################
	// Headerangaben pruefen

	// Eingetragene Strukturgroesse sicherstellen
	if (!header.sizeData) {
		header.sizeData = sizeof(SDFFileDeviceHeader);
	}

	// Dateiversionsangabe
	if (!CheckVersionFile(header)) {

		return FALSE;
	}

	// Firmwareversion pruefen
	if ((strlen(header.versionFirmware) < 8) || (strlen(header.versionFirmware) > 32)) {

		// Fehler zuweisen
		m_errorState = ES_FW_VERSION_LENGTH;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_FW_VERSION_LENGTH_TEXT);
		return FALSE;
	}

	// Builddatum, wenn nicht vorhanden, eintragen
	if (strlen(header.timestamp) != 19) {

		time_t ltime;
		struct tm *ltm;

		time(&ltime);
		ltm = localtime(&ltime);
		strftime(header.timestamp, sizeof(header.timestamp), "%Y-%m-%d %H:%M:%S", ltm);
	}

	// CRC32 fuer Program-Flash berechnen
	header.crcProgram = GetCRC32(GetController()->GetMemProgram(0), GetController()->GetProgramLastAddress() - GetController()->GetProgramFirstAddress() + 1);

	// CRC16 fuer Dateiheader berechnen
	unsigned char *pCRC = &((unsigned char*)&header)[header.sizeData - sizeof(unsigned short)];
	*(unsigned short*)pCRC = GetCRC16((unsigned char*)&header, header.sizeData - sizeof(unsigned short));

	// Header loeschen
	memset(m_headBuf, 0, sizeof(m_headBuf));
	m_header = (SDFFileDeviceHeader*)m_headBuf;
	// Struktur uebernehmen
	*m_header = header;

	
	//############################
	// Schreiben der Hexdatei

	FILE *fp;
	char data[0x400];
	unsigned short i;
	unsigned long numLine = 0;

	if ((fp = fopen(fileName, "w")) == (FILE*)0) {
		// Oeffnen der Datei gescheitert
		m_errorState = ES_FILE_OPEN;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_FILE_OPEN_TEXT);
		return FALSE;
	}

	// Dateiheader schreiben
	for (i=0; i < header.sizeData; i += INTELHEX_DATA_COUNT) {

		// Extended Segment Address Record
		if (!MakeTypedDataLine(i, INTELHEX_TYPE_DEVICE_HEADER, &m_headBuf[i],
			(header.sizeData - i) > INTELHEX_DATA_COUNT ? INTELHEX_DATA_COUNT : header.sizeData - i, data)) {

			// Fehler zuweisen
			m_errorState = ES_INTELHEX_MAKE_RECORD;
			sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_INTELHEX_MAKE_RECORD_TEXT);
			break;
		}

		fprintf(fp, "%s\n", data);
		numLine++;
	}

	// Abbruch nach Fehler
	if (m_errorState != ES_NO) {

		fclose(fp);
		return FALSE;
	}
	
	// Verschluesselungstabelle schreiben
	for (i=0; m_isCrypt && (i < sizeof(m_crypt)); i += INTELHEX_DATA_COUNT) {

		// Extended Segment Address Record
		if (!MakeTypedDataLine(i, INTELHEX_TYPE_DEVICE_CRYPT, &m_crypt[i],
			(sizeof(m_crypt) - i) > INTELHEX_DATA_COUNT ? INTELHEX_DATA_COUNT : sizeof(m_crypt) - i, data)) {

			// Fehler zuweisen
			m_errorState = ES_INTELHEX_MAKE_RECORD;
			sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_INTELHEX_MAKE_RECORD_TEXT);
			break;
		}

		fprintf(fp, "%s\n", data);
		numLine++;
	}
	fclose(fp);

	// Abbruch nach Fehler
	if (m_errorState != ES_NO) {

		return FALSE;
	}

	// Firmwaredaten schreiben
	if (!m_fileRawFirmware.Save(fileName, TRUE)) {

		return FALSE;
	}

	// Anzeigedaten in Datei schreiben
	if (m_fileDisplay.GetSizeData()) {
		if (!m_fileDisplay.Save(fileName, TRUE)) {

			return FALSE;
		}
	}

	// Geraetetexte
	if (m_fileDeviceText.GetSizeData()) {
		if (!m_fileDeviceText.Save(fileName, TRUE)) {

			return FALSE;
		}
	}

	// Endezeile eintragen
	if ((fp = fopen(fileName, "a")) == (FILE*)0) {
		// Oeffnen der Datei gescheitert
		m_errorState = ES_FILE_OPEN;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_FILE_OPEN_TEXT);
		return FALSE;
	}

	if (!MakeEndLine(data)) {
		
		fclose(fp);
		// Fehler zuweisen
		m_errorState = ES_INTELHEX_MAKE_RECORD;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_INTELHEX_MAKE_RECORD_TEXT);
		return FALSE;
	}

	fprintf(fp, "%s\n", data);
	numLine++;

	fclose(fp);

	return TRUE;
}

int CDFFileDevice::Encrypt(unsigned long numSerial)
{
	unsigned long address;
	unsigned long firstAddress = GetController()->GetProgramFirstAddress();
	unsigned long lastAddress  = GetController()->GetProgramLastAddress();
	unsigned char *mem = GetController()->GetMemProgram(0);

	if (m_isCrypt) {
		// Verschluesselungsveruch obwohl bereits verschluesselt!
		m_errorState = ES_CRYPT_IS;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_CRYPT_IS_TEXT);
		return FALSE;
	}

	srand(numSerial);

	for (address=0; address < IV_DFFILEDEVICE_SIZE_CRYPT; address++) {

		m_crypt[address] = (unsigned char)(rand() % 0x100);
	}

	m_crypt[IV_DFFILEDEVICE_SN_3] = (unsigned char)(numSerial >> 24);
	m_crypt[IV_DFFILEDEVICE_SN_2] = (unsigned char)(numSerial >> 16);
	m_crypt[IV_DFFILEDEVICE_SN_1] = (unsigned char)(numSerial >> 8);
	m_crypt[IV_DFFILEDEVICE_SN_0] = (unsigned char)(numSerial & 0xFF);

	m_crypt[IV_DFFILEDEVICE_CF_3] = (unsigned char)(IV_DFFILEDEVICE_CRYPT_FIX >> 24);
	m_crypt[IV_DFFILEDEVICE_CF_2] = (unsigned char)(IV_DFFILEDEVICE_CRYPT_FIX >> 16);
	m_crypt[IV_DFFILEDEVICE_CF_1] = (unsigned char)(IV_DFFILEDEVICE_CRYPT_FIX >> 8);
	m_crypt[IV_DFFILEDEVICE_CF_0] = (unsigned char)(IV_DFFILEDEVICE_CRYPT_FIX & 0xFF);

	for (address=firstAddress; address <= lastAddress; address++) {

		mem[address] ^= m_crypt[address % IV_DFFILEDEVICE_SIZE_CRYPT];
	}
	m_isCrypt = TRUE;

	return TRUE;
}

int CDFFileDevice::Decrypt()
{
	unsigned long cryptFix;
	unsigned long address;
	unsigned long firstAddress = GetController()->GetProgramFirstAddress();
	unsigned long lastAddress  = GetController()->GetProgramLastAddress();
	unsigned char *mem = GetController()->GetMemProgram(0);

	if (!m_isCrypt) {
		// Entschluesselungsversuch ohne vorherige Verschluesselung!
		m_errorState = ES_CRYPT_NONE;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_CRYPT_NONE_TEXT);
		return FALSE;
	}

	m_numSerial = m_crypt[IV_DFFILEDEVICE_SN_0];
	m_numSerial |= (unsigned long)m_crypt[IV_DFFILEDEVICE_SN_1] << 8;
	m_numSerial |= (unsigned long)m_crypt[IV_DFFILEDEVICE_SN_2] << 16;
	m_numSerial |= (unsigned long)m_crypt[IV_DFFILEDEVICE_SN_3] << 24;

	cryptFix = m_crypt[IV_DFFILEDEVICE_CF_0];
	cryptFix |= (unsigned long)m_crypt[IV_DFFILEDEVICE_CF_1] << 8;
	cryptFix |= (unsigned long)m_crypt[IV_DFFILEDEVICE_CF_2] << 16;
	cryptFix |= (unsigned long)m_crypt[IV_DFFILEDEVICE_CF_3] << 24;

	// Wert ist konstant und dient der Sicherung von der Verschluesselungstabelle
	if (cryptFix != IV_DFFILEDEVICE_CRYPT_FIX) {

		// Fehler zuweisen
		m_errorState = ES_CRYPT_FIXFAIL;
		sprintf(m_errorDetail, "%s(%i): %s", __FILE__, __LINE__, ES_CRYPT_FIXFAIL_TEXT);
		return FALSE;
	}

	for (address=firstAddress; address <= lastAddress; address++) {

		mem[address] ^= m_crypt[address % IV_DFFILEDEVICE_SIZE_CRYPT];
	}
	m_isCrypt = FALSE;

	return TRUE;
}

int CDFFileDevice::Build(char *fileDevice, SDFFileDeviceHeader &header, char *fileFirmware, char *fileDisplayData, char *fileDeviceText)
{
	// Umgebungsreset
	Reset();
	// Fehlerobjekt zuruecksetzen
	ResetError();


	// Laden der Firmware
	if (!m_fileRawFirmware.Load(fileFirmware)) {

		return FALSE;
	}
	if (strlen(fileDisplayData)) {
		// Laden der Anzeigedaten
		if (!m_fileDisplay.Load(fileDisplayData)) {

			return FALSE;
		}
	}
	if (strlen(fileDeviceText)) {
		
		// Importieren der Geraetetexte
		if (!m_fileDeviceText.Import(fileDeviceText)) {

			return FALSE;
		}
	}

	// Schreiben der Geraetedatei
	if (!Save(fileDevice, header)) {

		return FALSE;
	}

	return TRUE;
}

unsigned long CDFFileDevice::GetSizeData()
{
	return (GetController()->GetProgramLastAddress() - GetController()->GetProgramFirstAddress() + 1)
		+ m_fileDisplay.GetSizeData() + m_fileDeviceText.GetSizeData();
}

int CDFFileDevice::CompareInfos(TDFHardwareInfo hwInfo, TDFSoftwareInfo swInfo)
{
	unsigned char type;
	unsigned long bit;
	int res = 1;

	// Keine Hardwareinfo per Software gesetzt worden?
	if (hwInfo.version == 0) {
		
		// Dieses Betrifft lediglich die Geraete: PZE-/ZK-MasterIV
		
		// Je nach Geraetetyp Standardkonfiguration annehmen.
		// Die meisten Optionen werden durch die (04.00.23) Firmware ermittelt und gesetzt.
		switch(hwInfo.typeDevice) {
		case 0: //IV_VALUE_TYPE_DEVICE_PZE
			// Zusaetzlich Tastatur und Display sicherstellen
			hwInfo.typeDisplay = 1;//IV_VALUE_TYPE_DISPLAY_SMALL;
			hwInfo.typeKeypad  = 1;//IV_VALUE_TYPE_KEYPAD_9;
			break;
		case 1: //IV_VALUE_TYPE_DEVICE_ZK
			// Zusaetzlich Zutrittskontrolle sicherstellen
			hwInfo.componentRS485_1 = 1;
			hwInfo.typeExternalReader = 1;//IV_VALUE_TYPE_EXTERNAL_READER_GIS;
			break;
		}
		
		// Ist die ZK-Option in Verwendung?
		if (hwInfo.administrationFlags & HIAF_HAS_ZK) {
			
			// ZK-Option mit GIS Lesern.
			hwInfo.componentRS485_1 = 1;
			hwInfo.typeExternalReader = 1;
		}
	}

	// Geraetetyp pruefen
	if (((1 << hwInfo.typeDevice) & swInfo.device) == 0) {

		strcpy(m_errorText, "Der Geraetetyp wird nicht unterstuetzt.");
		return 0;
	}

	type = 0;
	// Boardtyp pruefen
	switch(hwInfo.typeBoard) {
	case 0x21: type++;
	case 0x20: type++;
	case 0x14: type++;
	case 0x13: type++;
	case 0x12: type++;
	case 0x11: type++;
	case 0x10:
		break;
	}
	if (((1 << type) & swInfo.board) == 0) {

		strcpy(m_errorText, "Die Boardrevision wird nicht unterstuetzt.");
		return 0;
	}

	// IO-Modul pruefen
	if (hwInfo.optionIOModule && !swInfo.iomodule) {

		strcpy(m_errorText, "Das IO-Modul wird nicht unterstuetzt.");
		return 0;
	}

	// Timeboyanbindung pruefen
	if (hwInfo.optionTBAdaption && !swInfo.docking) {

		strcpy(m_errorText, "Die Timeboyanbindung wird nicht unterstuetzt.");
		return 0;
	}

	// Comserver pruefen
	if (hwInfo.optionComserver && !swInfo.comserver) {

		strcpy(m_errorText, "Der Comserver-Modus wird nicht unterstuetzt.");
		return 0;
	}

	// Transponderwert schreiben pruefen
	if (hwInfo.optionTPWrite && !swInfo.transponder_write) {

		strcpy(m_errorText, "Das Schreiben eines Transponders wird nicht unterstuetzt.");
		return 0;
	}

	// Http-Protokoll pruefen
	if (hwInfo.optionHttp && !swInfo.http) {

		strcpy(m_errorText, "Das Http-Protokoll wird nicht unterstuetzt.");
		return 0;
	}

	// Sprachdatei pruefen
	if (hwInfo.optionLanguage && !swInfo.language) {

		strcpy(m_errorText, "Die Sprachdateien werden nicht unterstuetzt.");
		return 0;
	}

	// Update des Biokey pruefen
	if (hwInfo.optionBiokeyUpdate && !swInfo.biokey_update) {

		strcpy(m_errorText, "Das Update des Biokey wird nicht unterstuetzt.");
		return 0;
	}

	// RS485 pruefen
	if (hwInfo.componentRS485_0 && !swInfo.rs485) {

		strcpy(m_errorText, "Die RS485 - Kommunikation wird nicht unterstuetzt.");
		return 0;
	}

	// TCPIP pruefen
	if (hwInfo.typeTCPIP && (((1 << (hwInfo.typeTCPIP - 1)) & swInfo.tcpip)) == 0) {
//	if (hwInfo.componentTCPIP && !swInfo.tcpip) {

		strcpy(m_errorText, "Die TCPIP - Kommunikation wird nicht unterstuetzt.");
		return 0;
	}

	// ModemMCxx pruefen
	if (hwInfo.typeModemMC && (((1 << (hwInfo.typeModemMC - 1)) & swInfo.modem)) == 0) {

		strcpy(m_errorText, "Das verwendete Modem wird nicht unterstuetzt.");
		return 0;
	}

	// RS485 (intern) ermitteln
	if (hwInfo.componentRS485_1 && !swInfo.access) {

		strcpy(m_errorText, "Die Zutrittskontrolle wird nicht unterstuetzt.");
		return 0;
	}

	// Display pruefen
	if (hwInfo.typeDisplay && ((1 << (hwInfo.typeDisplay - 1)) & swInfo.display) == 0) {

		strcpy(m_errorText, "Das verwendete Display wird nicht unterstuetzt.");
		return 0;
	}

	// Tastatur pruefen
	if (hwInfo.typeKeypad && ((1 << (hwInfo.typeKeypad - 1)) & swInfo.keypad) == 0) {

		strcpy(m_errorText, "Die verwendete Tastatur wird nicht unterstuetzt.");
		return 0;
	}
	
	// Transponder pruefen
	if (hwInfo.typeTransponder && ((1 << (hwInfo.typeTransponder - 1)) & swInfo.transponder) == 0) {

		strcpy(m_errorText, "Der verwendete Transponderleser wird nicht unterstuetzt.");
		return 0;
	}

	// MMC pruefen
	if (hwInfo.typeMMC && ((1 << (hwInfo.typeMMC - 1)) & swInfo.mmc) == 0) {

		strcpy(m_errorText, "Die verwendete MMC - Option wird nicht unterstuetzt.");
		return 0;
	}

	// Funk pruefen
	if (hwInfo.typeRadio && ((1 << (hwInfo.typeRadio - 1)) & swInfo.bluetooth) == 0) {

		strcpy(m_errorText, "Das verwendete Funkmodul wird nicht unterstuetzt.");
		return 0;
	}

	// GPS pruefen
	if (hwInfo.typeGPS && ((1 << (hwInfo.typeGPS - 1)) & swInfo.gps) == 0) {

		strcpy(m_errorText, "Das verwendete GPS - Modul wird nicht unterstuetzt.");
		return 0;
	}

	// USB pruefen
	if (hwInfo.componentUSB && !swInfo.usb) {

		strcpy(m_errorText, "Die USB - Kommunikation wird nicht unterstuetzt.");
		return 0;
	}

	// DCF77 pruefen
	if (hwInfo.typeDCF77 && ((1 << (hwInfo.typeDCF77 - 1)) & swInfo.dcf77) == 0) {

		strcpy(m_errorText, "Das verwendete DCF77 - Modul wird nicht unterstuetzt.");
		return 0;
	}


	// DigitalIn pruefen
	if (hwInfo.componentInAnalog) {

		for (bit=1;; bit <<= 1) {

			if ((bit & hwInfo.componentInAnalog) && !(bit & swInfo.analog_in)) {

				strcpy(m_errorText, "Analoger Eingang wird nicht unterstuetzt.");
				return 0;
			}

			if (bit == 0x8000000) break;
		}
	}

	// DigitalIn pruefen
	if (hwInfo.componentInDigital) {

		for (bit=1;; bit <<= 1) {

			if ((bit & hwInfo.componentInDigital) && !(bit & swInfo.digital_in)) {

				strcpy(m_errorText, "Digitaler Eingang wird nicht unterstuetzt.");
				return 0;
			}

			if (bit == 0x8000000) break;
		}
	}

	// DigitalOut pruefen
	if (hwInfo.componentOutDigital) {

		for (bit=1;; bit <<= 1) {

			if ((bit & hwInfo.componentOutDigital) && !(bit & swInfo.digital_out)) {

				strcpy(m_errorText, "Digitaler Ausgang wird nicht unterstuetzt.");
				return 0;
			}

			if (bit == 0x8000000) break;
		}
	}

	// Fingerprint pruefen
	if (hwInfo.typeFingerprint && ((1 << (hwInfo.typeFingerprint - 1)) & swInfo.fingerprint) == 0) {

		strcpy(m_errorText, "Das verwendete Fingerprint - Modul wird nicht unterstuetzt.");
		return 0;
	}

	// Externe Zutrittsleser pruefen
	if (hwInfo.typeExternalReader && ((hwInfo.typeExternalReader & swInfo.access) == 0)) {

		strcpy(m_errorText, "Die Externen Zutrittsleser werden nicht unterstuetzt.");
		return 0;
	}

	// Sollen die Setupoptionen geprueft werden
	if (hwInfo.administrationFlags & HIAF_HAS_OPTIONS3) {

		// SetupOption pruefen
		if (hwInfo.optionsSetup) {

			for (bit=1;; bit <<= 1) {

				if ((bit & hwInfo.optionsSetup) && !(bit & swInfo.setup_options)) {

					strcpy(m_errorText, "Softwareoption wird nicht unterstuetzt.");
					return 0;
				}

				if (bit == 0x8000000) break;
			}
		}

	}

	return res;
}

int CDFFileDevice::GetInfos(TDFSoftwareInfo swInfo, char *info)
{
	char tmp[0x100];
	unsigned long bit;
	int res = 1;

	memset(tmp, 0, sizeof(tmp));
	strcpy(tmp, "Unterstuetzte Optionen/Module\r\n");
	strcat(tmp, "------------------------------------------\r\n");
	strcat(info, tmp);

	// SofwareInfo verfuegbar?
	if (!swInfo.size) {

		return 0;
	}

	memset(tmp, 0, sizeof(tmp));
	strcat(tmp, "\r\nBasis\r\n----------\r\n");
	strcat(info, tmp);

	strcat(info, "Geraete: ");
	memset(tmp, 0, sizeof(tmp));
	// Moegliche Geraetetypen
	for (bit=1; ; bit <<= 1) {

		if (bit & swInfo.device) {

			if (strlen(tmp)) {
				strcat(tmp, ", ");
			}

			switch(bit) {
			case DEVICE_PZE:     strcat(tmp, "PZE-MasterIV"); break;
			case DEVICE_ZK:      strcat(tmp, "ZK-MasterIV");  break;
			case DEVICE_BDE:     strcat(tmp, "BDE-MasterIV"); break;
			case DEVICE_MOBIL:   strcat(tmp, "Mobil-MasterIV"); break;
			case DEVICE_MDEBOX:  strcat(tmp, "MDE-BoxIV"); break;
			case DEVICE_AE:      strcat(tmp, "AE-MasterIV"); break;
			case DEVICE_TIMEBOY: strcat(tmp, "Timeboy-MasterIV"); break;
			default:             strcat(tmp, "Undefiniert"); break;
			}
		}

		if (bit == 0x80) break;;
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	strcat(info, "Platinen: ");
	memset(tmp, 0, sizeof(tmp));
	// Boardtyp pruefen
	for (bit=1; ; bit <<= 1) {

		if (bit & swInfo.board) {

			if (strlen(tmp)) {
				strcat(tmp, ", ");
			}

			switch(bit) {
			case BOARD_10: strcat(tmp, "1.0"); break;
			case BOARD_11: strcat(tmp, "1.1"); break;
			case BOARD_12: strcat(tmp, "1.2"); break;
			case BOARD_13: strcat(tmp, "1.3"); break;
			case BOARD_14: strcat(tmp, "1.4"); break;
			case BOARD_20: strcat(tmp, "2.0"); break;
			case BOARD_21: strcat(tmp, "2.1"); break;
			default:       strcat(tmp, "Undefiniert"); break;
			}
		}

		if (bit == 0x80) break;
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	/////////////////////////////////////////////
	// Hauptkommunikationen
	memset(tmp, 0, sizeof(tmp));
	strcat(tmp, "\r\nHauptkommunikationen\r\n------------------------------\r\n");
	strcat(info, tmp);
	
	// Funk pruefen
	strcat(info, "Bluetooth: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.bluetooth) {
		for (bit=1; ; bit <<= 1) {

			if (bit & swInfo.bluetooth) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) {
				case BLUETOOTH_WRAPTHOR: strcat(tmp, "Ja"); break;
				default:                 strcat(tmp, "Undefiniert"); break;
				}

			}

			if (bit == 0x80) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	// ModemMCxx pruefen
	strcat(info, "Mobil-Funkmodem: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.modem) {
		for (bit=1; ; bit <<= 1) {

			if (bit & swInfo.modem) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) {
				case MODEM_MC39: strcat(tmp, "Siemens MC39"); break;
				case MODEM_MC55: strcat(tmp, "Siemens MC55"); break;
				default:         strcat(tmp, "Undefiniert"); break;
				}
			}

			if (bit == 0x80) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	// RS485 pruefen
	sprintf(tmp, "RS485 Kommunikation: %s\r\n", swInfo.rs485 ? "Ja" : "Nein");
	strcat(info, tmp);

	// TCPIP pruefen
	sprintf(tmp, "TCP/IP Kommunikation: %s\r\n", swInfo.tcpip ? "Ja" : "Nein");
	strcat(info, tmp);

	// USB pruefen
	sprintf(tmp, "USB Kommunikation: %s\r\n", swInfo.usb ? "Ja" : "Nein");
	strcat(info, tmp);

	/////////////////////////////////////////////
	// Zusatzoptionen
	memset(tmp, 0, sizeof(tmp));
	strcat(tmp, "\r\nZustzliche Optionen der Firmware\r\n------------------------------\r\n");
	strcat(info, tmp);

	// IOModule
	sprintf(tmp, "IO-Modul: %s\r\n", swInfo.iomodule ? "Ja" : "Nein");
	strcat(info, tmp);

	// Timboyanbindung
	sprintf(tmp, "Timeboyanbindung: %s\r\n", swInfo.docking ? "Ja" : "Nein");
	strcat(info, tmp);

	// Untersttzung COM-Server
	sprintf(tmp, "COM-Server: %s\r\n", swInfo.comserver ? "Ja" : "Nein");
	strcat(info, tmp);

	// Transpondermen
	sprintf(tmp, "Transpondermen: %s\r\n", swInfo.transponder_write ? "Ja" : "Nein");
	strcat(info, tmp);

	// BiokeyUpdate
	sprintf(tmp, "Biokeyupdate: %s\r\n", swInfo.biokey_update ? "Ja" : "Nein");
	strcat(info, tmp);

	// Setupoptionen (intern) ermitteln
	strcat(info, "Setupoptionen: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.setup_options) {
		for (bit=1; ; bit <<= 1) {

			if (bit & swInfo.setup_options) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) {
				case SETUP_OPTION_IOCONTROL:     strcat(tmp, "Signalverarbeitung"); break;
				default:             strcat(tmp, "Undefiniert"); break;
				}
			}

			if (bit == 0x80) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	/////////////////////////////////////////////
	// Zusatzbaugruppen
	memset(tmp, 0, sizeof(tmp));
	strcat(tmp, "\r\nZusatzbaugruppen\r\n------------------------------\r\n");
	strcat(info, tmp);

	// Display pruefen
	strcat(info, "Anzeige: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.display) {
		for (bit=1; ; bit <<= 1) {

			if (bit & swInfo.display) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) {
				case DISPLAY_LEHNER_SWAP: strcat(tmp, "4 Zoll Monochrome"); break;
				case DISPLAY_LEHNER:      strcat(tmp, "6 Zoll Monochrome"); break;
				case DISPLAY_NEUMUELLER:  strcat(tmp, "1,9 Zoll Color"); break;
				default:                  strcat(tmp, "Undefiniert"); break;
				}
			}

			if (bit == 0x80) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	// DigitalOut pruefen
	strcat(info, "Ausgaenge Digital: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.digital_out) {

		for (bit=1;; bit <<= 1) {

			if (bit & swInfo.digital_out) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) { // Digitale Ausgaenge
				case 0x01: strcat(tmp, "dO1"); break;
				case 0x02: strcat(tmp, "dO2"); break;
				case 0x04: strcat(tmp, "dO3"); break;
				case 0x08: strcat(tmp, "dO4"); break;
				default:   strcat(tmp, "Undefiniert"); break;
				}
			}

			if (bit == 0x8000000) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	// AnalogIn pruefen
	strcat(info, "Eingaenge Analog: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.analog_in) {

		for (bit=1;; bit <<= 1) {

			if (bit & swInfo.analog_in) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) { // Analoge Eingaenge
				case 0x01: strcat(tmp, "aI1"); break;
				case 0x02: strcat(tmp, "aI2"); break;
				default:   strcat(tmp, "Undefiniert"); break;
				}
			}

			if (bit == 0x8000000) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	// DigitalIn pruefen
	strcat(info, "Eingaenge Digital: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.digital_in) {

		for (bit=1;; bit <<= 1) {

			if (bit & swInfo.digital_in) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) { // Digitale Eingaenge
				case 0x01: strcat(tmp, "dI1"); break;
				case 0x02: strcat(tmp, "dI2"); break;
				case 0x04: strcat(tmp, "dI3"); break;
				case 0x08: strcat(tmp, "dI4"); break;
				case 0x10: strcat(tmp, "dI5"); break;
				case 0x20: strcat(tmp, "dI6"); break;
				case 0x40: strcat(tmp, "dI7"); break;
				case 0x80: strcat(tmp, "dI8"); break;
				default:   strcat(tmp, "Undefiniert"); break;
				}
			}

			if (bit == 0x8000000) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	// Fingerprint pruefen
	strcat(info, "Fingerprint: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.fingerprint) {
		for (bit=1; ; bit <<= 1) {

			if (bit & swInfo.fingerprint) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) {
				case FP_BIOKEY: strcat(tmp, "Ja"); break;
				default:        strcat(tmp, "Undefiniert"); break;
				}
			}

			if (bit == 0x80) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);
	
	// DCF77 pruefen
	strcat(info, "Funkuhr: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.dcf77) {
		for (bit=1; ; bit <<= 1) {

			if (bit & swInfo.dcf77) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) {
				case DCF77_HKW:  strcat(tmp, "Ja"); break;
				default:         strcat(tmp, "Undefiniert"); break;
				}

			}

			if (bit == 0x80) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	// GPS pruefen
	strcat(info, "GPS: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.gps) {
		for (bit=1; ; bit <<= 1) {

			if (bit & swInfo.gps) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) {
				case GPS_FURUNO: strcat(tmp, "Ja"); break;
				default:         strcat(tmp, "Undefiniert"); break;
				}

			}

			if (bit == 0x80) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	// MMC pruefen
	strcat(info, "MM Card: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.mmc) {
		for (bit=1; ; bit <<= 1) {

			if (bit & swInfo.mmc) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) {
				case MMC_MEMORY: strcat(tmp, "Speichererweiterung"); break;
				case MMC_DISK:   strcat(tmp, "Wechseldatentrger"); break;
				case MMC_USBHOST:strcat(tmp, "USB-Memory-Stick"); break;
				default:         strcat(tmp, "Undefiniert"); break;
				}
			}

			if (bit == 0x80) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	// Tastatur pruefen
	strcat(info, "Tastatur: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.keypad) {
		for (bit=1; ; bit <<= 1) {

			if (bit & swInfo.keypad) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) {
				case KEYPAD_PZE:       strcat(tmp, "9 Tasten"); break;
				case KEYPAD_BDE:       strcat(tmp, "58 Tasten"); break;
				case KEYPAD_AE:        strcat(tmp, "29 Tasten"); break;
				case KEYPAD_PZE_BASIC: strcat(tmp, "7 Tasten"); break;
				case KEYPAD_TIMEBOY:   strcat(tmp, "25 Tasten"); break;
				default:               strcat(tmp, "Undefiniert"); break;
				}
			}

			if (bit == 0x80) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	// Transponder pruefen
	strcat(info, "Transponderleser: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.transponder) {
		for (bit=1; ; bit <<= 1) {

			if (bit & swInfo.transponder) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) {
				case TP_TSR30:       strcat(tmp, "[EM4102, Hitag2]"); break;
				case TP_TSR32:       strcat(tmp, "[EM4102, EM4450, Hitag]"); break;
				case TP_PRIMO100:    strcat(tmp, "[Legic]"); break;
				case TP_MIFARE:      strcat(tmp, "[Mifare]"); break;
				case TP_IBUTTON_EXT: strcat(tmp, "[IButton (Extern)]"); break;
				case TP_SMARTRELAIS: strcat(tmp, "[SimonsVoss]"); break;
				case TP_NEDAP:       strcat(tmp, "[Nedap]"); break;
				case TP_IBUTTON_INT: strcat(tmp, "[IButton (Intern)]"); break;
				default:             strcat(tmp, "[Undefiniert]"); break;
				}
			}

			if (bit == 0x80) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);
	
	// RS485 (intern) ermitteln
	strcat(info, "Zutrittskontrolle: ");
	memset(tmp, 0, sizeof(tmp));
	if (swInfo.access) {
		for (bit=1; ; bit <<= 1) {

			if (bit & swInfo.access) {

				if (strlen(tmp)) {
					strcat(tmp, ", ");
				}

				switch(bit) {
				case ACCESS_GIS:     strcat(tmp, "TS-Serie"); break;
				case ACCESS_PHG:     strcat(tmp, "PHG-Serie"); break;
				case ACCESS_PROTECT: strcat(tmp, "Protect-Serie"); break;
				default:             strcat(tmp, "Undefiniert"); break;
				}
			}

			if (bit == 0x80) break;
		}
	} else {
		strcat(tmp, "Nein");
	}
	strcat(tmp, "\r\n");
	strcat(info, tmp);

	return res;
}

